#
# Virtual Environments, Reykjavik University
# Lab 16th of February, 2007
# hannes@ru.is
#
# Roaming Ralph with nice mouse controls and proper collisions.
# Original code from Cyan on the Panda 3D forums
# (Posted Tue Oct 31, 2006)
#
from pandac import PandaModules as P #alias PandaModules as P
import direct.directbase.DirectStart # Start Panda
from direct.showbase.DirectObject import DirectObject # To listen for Events
from direct.task import Task # To use Tasks
from direct.actor import Actor # To use animated Actors
from direct.interval import LerpInterval as LERP#alias LerpInterval as LERP
from direct.fsm import FSM# To use Finite State Machines.
from direct.showbase.PythonUtil import clampScalar#useful.
from keybindings import Controls
from pandac.PandaModules import PandaNode,LightNode,TextNode
from pandac.PandaModules import Filename
from pandac.PandaModules import NodePath
from pandac.PandaModules import Point3,Vec4
from direct.task.Task import Task
from direct.showbase.BufferViewer import BufferViewer
from pandac.PandaModules import CardMaker
import sys,os
from direct.gui.OnscreenText import OnscreenText

base.cTrav = P.CollisionTraverser()#initialize traverser
#Collision detection can fail if objects move too fast. (Quantum tunnelling.)
base.cTrav.setRespectPrevTransform(1)#fluid move prevents quantum tunnelling.
#Global collision handlers.
Pusher = P.CollisionHandlerPusher()#Pusher keeps its ward out of things.
Floor = P.CollisionHandlerFloor()#Floor keeps its ward grounded.
#collision masks. Use a bitwise or (operator | ) if more than one apply.
AllM = P.BitMask32.bit(0)#for everything in the world, except display elements
GroundM = P.BitMask32.bit(1)#the ground. What the floor rays collide with.
PusherM = P.BitMask32.bit(2)#Solid objects, but not the ground.
CameraM = P.BitMask32.bit(3)#What shouldn't be between camera and avatar.

#setup keyboard and mouse
con = Controls(translation={#initialize global controls
    'zoom in':['wheel_up','arrow_up'],
    'zoom out':['wheel_down','arrow_down'],
    'click':['mouse1','mouse3']})
    
def fromCol(parent,handler,type,mask = P.BitMask32.allOn()):
        """Setup a from collision solid.    
        Last I checked CollisionPolygon 's and CollisionTube 's can't be used
        as from solids. If you pass one, it won't hit anything"""
        nodepath = parent.attachNewNode(P.CollisionNode('frmcol'))
        nodepath.node().addSolid(type)#add the solid to the new collisionNode
        nodepath.node().setFromCollideMask(mask)#allow selective masking
        nodepath.setCollideMask(P.BitMask32.allOff())#it's a from solid only.
        ####uncomment this line to make the collision solid visible:
        ##nodepath.show()
        base.cTrav.addCollider(nodepath,handler)#add to the traverser
        try:#the next line doesn't work on queues. (not necessary)
            handler.addCollider(nodepath,parent)#keep the ward out of trouble
        except:
            pass#Don't care. This method needs to work on queues too.
        return nodepath#we might need the new CollisionNode again later.

class HealthBar(NodePath):
    """ Adapted from drwr's code on the Panda 3D forums """
    
    def __init__(self):
        NodePath.__init__(self, 'healthbar')
        ##self.setTransparency(P.TransparencyAttrib.MAlpha)
        self.group = self.attachNewNode('group')
        self.group.setPos(-0.5,0,0)
        cmfg = CardMaker('fg')
        cmfg.setFrame(0, 1, -0.1, 0.1)
        self.fg = self.group.attachNewNode(cmfg.generate())
        cmbg = CardMaker('bg')
        cmbg.setFrame(-1, 0, -0.1, 0.1)
        self.bg = self.group.attachNewNode(cmbg.generate())
        self.bg.setPos(1, 0, 0)
        self.fg.setColor(0, 1, 0, 0.5)
        self.bg.setColor(0.5, 0.5, 0.5, 0.5)
        self.value = 0.8
        self.update()
        self.setBillboardAxis()
        
    def update(self):
        self.fg.setScale(self.value, 1, 1)
        self.bg.setScale(1.0 - self.value, 1, 1)        

    def setHealth(self, value):
        self.value = value
        self.update()
        
    def decrease(self, delta):
        if (self.value - delta) >= 0:
            self.value -= delta
            self.update()
        
class OverheadMap(NodePath):
    
    def __init__(self, x1, y1, x2, y2):
        """ Create a square map with the given world corners in the x-y plane """
        NodePath.__init__(self, 'overheadmap')
        self.setBoundary(x1,y1,x2,y2)
                
        cmbg = CardMaker('bg')
        cmbg.setFrame(0,1,0,1)
        self.prime = self.attachNewNode(cmbg.generate())
        self.prime.setColor(0.5,0.5,0.5,0.5)
        
        cmav = CardMaker('am')
        cmav.setFrame(-0.02,0.02,-0.02,0.02)
        self.avatar = self.prime.attachNewNode(cmav.generate())
        self.avatar.setPos(0,0,0)
        self.avatar.setColor(0,1,0,1)  
        
        self.lm = []
            
    def addLandmarks(self, landmarks, color=P.Vec4(1,0,0,0.5)):
        """ Take in a list of (x,y) touples in world coordinates and place them
            as little dots on the map in the given color """
        cmlm = CardMaker('lm')
        cmlm.setFrame(-0.02,0.02,-0.02,0.02)
        for landmark in landmarks:
            lmpn = self.prime.attachNewNode(cmlm.generate())
            lmpn.setPos((landmark[0]-self.xoffset)*self.xscale,0,(landmark[1]-self.yoffset)*self.yscale)
            lmpn.setColor(color)
            self.lm.append(lmpn)
        
    def setBoundary(self, x1, y1, x2, y2):
        """ Define the extent of the map in world coordinates.  Create the proper
            offset and scaling values that transform world coordinates to map coordinates"""
        self.xoffset = x1
        self.xscale = 1.0/(x2-x1)
        self.yoffset = y1
        self.yscale = 1.0/(y2-y1)
        
    def setAvatarPos(self, x, y):
        """ Sets the position of the avatar marker to the given world coordinates """
        self.avatar.setPos((x-self.xoffset)*self.xscale,-0.01,(y-self.yoffset)*self.yscale)
            

class Box(DirectObject):
    
    def __init__(self, root, x, y):
        self.prime = loader.loadModel("Models/box")
        self.prime.reparentTo(root)
        self.prime.setPos(x,y,4)
        self.prime.setScale(1)
        self.prime.setCollideMask(AllM|PusherM|CameraM)

class Avatar(FSM.FSM):
    """ This is the character that represents the user in the environment """

    def __init__(self,iniHight=5,dest=P.Point3.zero(),speed=3,act={},solid={}):
        """Setup Avatar.  The default values load Ralph, but this can be easily changed."""

        #initialize movement variables
        self.speed = speed          #moving speed
        self.point = dest           #destination
        self.vel = P.Vec3.zero()    #velocity
        #You must call FSM init if you override init.
        FSM.FSM.__init__(self,'avatar')
        #Avatar scenegraph setup.
        self.prime=P.NodePath('avatar prime')   #prime: Avatar's primary nodepath.
        self.prime.reparentTo(render)           #Make Avatar visible.
        self.prime.setZ(iniHight)               #Be sure to start above the floor.
        self.__initActor(**act)                 #unpacks act dictionary as optional params
        self.__initSolids(**solid)              #same with solid
        
        self.health = HealthBar()
        self.health.setZ(1.2)
        self.health.reparentTo(self.prime)
        
        self.map = None
        
        #default to standing state instead of off.
        self.request('Stand')
    

    def __initActor(self,model='models/ralph',
            hprs=(180,0,0,.2,.2,.2),#ralph's Y is backward and he's too big.
            anims={"run":"models/ralph-run","walk":"models/ralph-walk"}):
        """Only seperate for organisation, treat it as is part of __init__() .
        Load actor and animations. Set coordinate offset from prime."""
        self.myActor=Actor.Actor(model,anims)
        self.myActor.setHprScale(*hprs)
        self.myActor.reparentTo(self.prime)#parent actor to the prime
        
    def __initSolids(self,ray=(0, 0, 1.3, 0, 0, -1),sphere=(0,0,.7,.4)):
        """Only seperate for organisation, treat it as is part of __init__() .       
        Set collision solids for the avatar."""
        #Ralph will have collision solids, so the
        # visible geometry shouldn't collide at all.
        self.prime.setCollideMask(P.BitMask32.allOff())
        # to collide with the terrain & keep ralph grounded
        fromCol(self.prime,Floor,P.CollisionRay(*ray),AllM|GroundM)
        #The ground and walls should have separate collision masks otherwise
        # Pusher can interfere with Floor if the ground gets too steep.
        #The sphere is set to collide with the other things in
        # the world which have mask PusherM
        fromCol(self.prime,Pusher,P.CollisionSphere(*sphere),PusherM)
        
    def setOverheadMap(self, map):
        self.map = map
        
    def update(self,dt):
        """Call this method in your main loop task."""
        self.prime.setFluidPos(#fluidly update position based on velocity
            self.prime.getX()+self.vel.getX()*dt,
            self.prime.getY()+self.vel.getY()*dt,
            self.prime.getZ() )#let Floor worry about Z
        #recalc velocity to point to destination
        self.vel = self.point-self.prime.getPos()
        if self.vel.lengthSquared()<.1:#Then consider yourself arrived.
            self.request('Stand')#change FSM state
            self.vel=P.Vec3.zero()#stop moving.
        else:
            self.vel.normalize()#set magnitude to 1
            self.vel*=self.speed#the magnitude of velocity is speed.
            if self.map is not None:
                self.map.setAvatarPos(self.prime.getX(),self.prime.getY())
        if self.state is "Run":
            self.health.decrease(0.001)
            
    def setDestination(self,point):
        """Go to the point."""
        self.point = point
        pr = self.prime.getP(),self.prime.getR()#to preserve pitch and roll
        self.prime.lookAt(self.point)#lookAt affects all three (HPR)
        #keep heading but revert pitch and roll.
        self.prime.setHpr(self.prime.getH(),*pr)
        #subtracting points yeilds a vector pointing from the 1st to the 2nd
        self.vel = self.point-self.prime.getPos()#destination from position
        self.vel.normalize()#set magnitude to 1
        self.vel*=self.speed#the magnitude of velocity is speed.
        self.request('Run')#change FSM state

## State handlers. Only define them if something happens at that transition.
    def enterRun(self):
        self.myActor.loop("run")#loop the run animation
        
    def exitRun(self):
        self.myActor.stop("run")#stop the run animation
    def enterStand(self):
        self.myActor.pose("walk",6)#both feet on the floor
        self.health.setHealth(1.0)
        ##print str(self.prime.getX())+"-"+str(self.prime.getY())
    #notice that no def exitStand is required.
    
#end Avatar

class Environment:
    """ The terrain and various features on it """
    
    def __init__(self):
        """Setup Environment
        Loads the model and sets collision masks."""
        self.prime = loader.loadModel("models/world")
        self.prime.reparentTo(render)
        #break the model into the named pieces
        for word in ('terrain','tree','wall','rock','hedge'):
            setattr(self,word,self.prime.find('**/%s*'%word))#isn't python fun?
        #set the collideMasks of the pieces individually
        self.terrain.setCollideMask(AllM|GroundM|CameraM)#Ground and walls \
        self.wall.setCollideMask(AllM|PusherM|CameraM)#should never be between\
        self.rock.setCollideMask(AllM|PusherM|CameraM)#camera and avatar.
        #But it doesn't matter if a tree is in the way. Let the player worry\
        self.tree.setCollideMask(AllM|PusherM)#about that. Besides, geom test\
        self.hedge.setCollideMask(AllM|PusherM)#for doodads is too expensive.
        
#end Environ

class Marker:
    """Setup marker to mark the clicked destination.
    At the moment simply loads jack and turns off his collisions, but you
    could easily make it do so much more."""
    def __init__(self):
        #you probably want to use a different marker model
        self.prime = loader.loadModel('jack')
        self.prime.reparentTo(render)
        self.prime.setScale(.1,.1,.1)
        #this is just a display element, so it shouldn't affect the world.
        self.prime.setCollideMask(P.BitMask32.allOff())#no collisions!
#end Marker

class EdgeScreenTracker(DirectObject):
    """Mouse camera control interface."""
    
    def __init__(self,avatar,offset=P.Point3.zero(), dist=10,
          rot=20,zoom=(2,30),pitch=(-80,-10)):
        # Disable default camera interface.
        base.disableMouse()
        # Set parameters
        self.zoomLvl = dist #camera starting distance
        self.speed = 1.0/rot # Controls speed of camera rotation.
        self.zoomClamp=zoom#clamp zoom in this range
        self.clampP=pitch#clamp pitch in this range
        self.target = avatar.attachNewNode('camera target')
        self.target.setPos(offset)#offset target from avatar.
        #Load the camera
        self.__loadCamera()
        #Enable new camera interface
        self.accept('zoom in', self.cameraZoom,[0.7])#Translated. For zooming.
        self.accept('zoom out', self.cameraZoom,[1.3])
        taskMgr.add(self.mousecamTask, "mousecamTask")#For edge screen tracking
        
    def __loadCamera(self):
        """Only seperate for organisation, treat it as is part of __init__() .
        Load the camera & setup segmet & queue for detecting obstructions."""
        #Don't rotate the target with the avatar.
        self.target.node().setEffect(P.CompassEffect.make(render))
        camera.reparentTo(self.target)# Attach the camera to target.
        camera.setPos(0,-self.zoomLvl,0)# Position the camera
        self.rotateCam(P.Point2(0,0))# Initialize gimbal clamps.
        self.Q = P.CollisionHandlerQueue()# New queue for camera.
        self.segment = fromCol(self.target,self.Q,
            P.CollisionSegment(P.Point3.zero(),camera.getPos(self.target)),
            P.BitMask32(CameraM))#CameraM into segment between camera & target.
            
    def mousecamTask(self,task):
        """Rotate camera when the pointer moves to the edges of the screen.
        Also temporarily zooms in past an obstructing CameraM'ed object."""
        self.setDist(self.zoomLvl)#preset dist to current zoom level.
        if self.Q.getNumEntries() > 0:#if there was a collision
            self.Q.sortEntries() #so we get the closest collision to avatar
            point=self.Q.getEntry(0).getSurfacePoint(self.target)#get the point
            if point.lengthSquared()<camera.getPos().lengthSquared():#not out.
                self.setDist(point.length())#Temporarily zoom to point.
        camera.lookAt(self.target)# always point camera at target
        if not base.mouseWatcherNode.hasMouse():#See if the mouse is available.
            return Task.cont#if no, just loop again.
        # Get the relative mouse position, its always between 1 and -1
        mpos = base.mouseWatcherNode.getMouse()
        if mpos.getX() > 0.8:
            self.rotateCam(P.Point2(-10,0))
        elif mpos.getX() < -0.8:
            self.rotateCam(P.Point2(10,0))
        if mpos.getY() > 0.9:
                self.rotateCam(P.Point2(0,-3))
        elif mpos.getY() < -0.9:
            self.rotateCam(P.Point2(0,3))
        return Task.cont#loop again.
    
    def rotateCam(self,arc):
        """Setup a lerp interval to rotate the camera about the target."""
        newP=clampScalar(self.target.getP()-arc.getY(),*self.clampP)#Clamped.
        newH=self.target.getH()+arc.getX()#Not clamped, just added.
        LERP.LerpHprInterval(self.target, self.speed,#Setup the interval\
            P.Vec3(newH,newP,self.target.getR(), ), ).start()#and start it.
            
    def cameraZoom(self,zoomFactor,):
        """Scale and clamp zoom level, then set distance by it."""
        self.zoomLvl=clampScalar(self.zoomLvl*zoomFactor,*self.zoomClamp)
        self.setDist(self.zoomLvl)
        
    def setDist(self,newDist):
        """Set camera distance from the target."""
        vec = camera.getPos()
        vec.normalize()#set length to 1
        vec*=newDist#set length to clamped value
        camera.setFluidPos(vec)#move the camera to new distance
        #Move the segment end but keep it a little behind and below the camera.
        self.segment.node().getSolid(0).setPointB(
            self.target.getRelativePoint(camera, P.Point3(0,-2,-1)))
#end EdgeScreenTracker

boxes = [(-10,-10),(20,40),(-60,-40),(-100,0),(-20,20),(10,30)]

class World(DirectObject):
    
    def __init__(self):
        #self.__setupShader()
        self.avatar = Avatar()#setup ralph
        self.environ = Environment().prime#load environment
        self.marker = Marker().prime#load marker
        
        aspect2d.setTransparency(P.TransparencyAttrib.MAlpha)
        self.map = OverheadMap(-128,-68,48,48)
        self.map.prime.setPos(-1*base.getAspectRatio(),0,-1)
        self.map.prime.setScale(0.5)
        self.map.prime.reparentTo(aspect2d)
        self.map.addLandmarks(boxes)
        
        self.avatar.setOverheadMap(self.map)
        
        for b in boxes:
            Box(render,b[0],b[1])
        
        EdgeScreenTracker(self.avatar.prime,P.Point3(0,0,1))#setup camera
        self.__setupPicking()
        self.last = 0#for calculating dt in gameLoop
        taskMgr.add(self.gameLoop, "gameLoop")#start the gameLoop task

    def __setupPicking(self):
        """Only seperate for organisation, treat it as is part of __init__() .
        Setup ray and queue for picking."""
        self.pickerQ = P.CollisionHandlerQueue()#the handler
        self.picker=fromCol(camera,self.pickerQ,P.CollisionRay(),GroundM)
        self.accept('click',self.OnClick)#translated
        ####uncomment this line to show the collisions:
        ##base.cTrav.showCollisions(render)
    
    def OnClick(self):
        """Handle the click event."""
        mpos=base.mouseWatcherNode.getMouse()#mouse's screen coordinates
        #This makes the ray's origin the camera and makes the ray point to mpos
        self.picker.node().getSolid(0).setFromLens(
            base.camNode,mpos.getX(),mpos.getY())
        #We don't want to traverse now, so wait for panda to do it, then move.
        taskMgr.doMethodLater(.02,self.__setDestination,'setDest')
        
    def __setDestination(self,task):
        #find the position of the nearest intersection in renderspace
        if self.pickerQ.getNumEntries() > 0:
            self.pickerQ.sortEntries() #this is so we get the closest object
            self.point=self.pickerQ.getEntry(0).getSurfacePoint(render)
            self.marker.setPos(self.point)#marker indicates destination
            self.avatar.setDestination(self.point)#set the destination
            
    def gameLoop(self,task):
        """The main game loop."""
        dt = task.time - self.last#time elapsed since the last frame.
        self.last = task.time#for calculating dt the next time
        self.avatar.update(dt)#allow avatar to update itself
        return direct.task.Task.cont#task must return cont to continue a loop
    
#end World

props = base.win.getProperties()
props.setFullscreen(True)
##props.setSize(800,800)
base.openMainWindow(props = props)

World()#no need to store to a variable
run() 